"""
This file defines the Karel class, which provides the actual
implementation of all functions described in the Karel Reference
Guide.

All instances of a Karel object store a reference to the world
in which they exist. Each Karel object exists on a given
(avenue, street) intersection and holds a certain number of beepers
in its beeper bag.


Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""
from .karel_ascii import compare_output, karel_ascii
from .karel_definitions import COLOR_MAP, INFINITY, Direction, KarelException
from .karel_world import KarelWorld

NEXT_DIRECTION_MAP = {
    Direction.NORTH: Direction.WEST,
    Direction.WEST: Direction.SOUTH,
    Direction.SOUTH: Direction.EAST,
    Direction.EAST: Direction.NORTH,
}
NEXT_DIRECTION_MAP_RIGHT = {v: k for k, v in NEXT_DIRECTION_MAP.items()}

# This map associates directions with the delta that Karel
# undergoes if it were to move one step in that direction
# delta is in terms of (avenue, street)
DIRECTION_DELTA_MAP = {
    Direction.NORTH: (0, 1),
    Direction.EAST: (1, 0),
    Direction.SOUTH: (0, -1),
    Direction.WEST: (-1, 0),
}


class Karel:
    def __init__(self, world_file=None):
        """
        This functions instantiates a new Karel instance and sets its
        location and current number of beepers to be the default starting
        values as indicated by the given world object.

        Parameters:
            world (KarelWorld) - The world that Karel should exists in

        Returns: None
        """
        self.world = KarelWorld(world_file)
        self._avenue, self._street = self.world.karel_starting_location
        self._direction = self.world.karel_starting_direction
        self._num_beepers = self.world.karel_starting_beeper_count

    def __repr__(self):
        return karel_ascii(self.world, self.street, self.avenue)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def compare_with(self, other):
        """
        Options:
            two_columns: bool (default=True)
        """
        if self == other:
            return True
        print(compare_output(self, other))
        return False

    @property
    def avenue(self):
        """
        This property can be used to access Karel's current avenue location.

        Parameters: None
        Returns:
            avenue (int) - The current avenue Karel is standing on.
        """
        return self._avenue

    @avenue.setter
    def avenue(self, val):
        """
        This property can be used to set Karel's current avenue location.

        Parameters:
            val (int) - The new avenue that Karel will be standing on.
        Returns: None
        """
        self._avenue = val

    @property
    def street(self):
        """
        This property can be used to access Karel's current street location.

        Parameters: None
        Returns:
            street (int) - The current street Karel is standing on.
        """
        return self._street

    @street.setter
    def street(self, val):
        """
        This property can be used to set Karel's current street location.

        Parameters:
            val (int) - The new street that Karel will be standing on.
        Returns: None
        """
        self._street = val

    @property
    def direction(self):
        """
        This property can be used to access Karel's current direction.

        Parameters: None
        Returns:
            street (Direction[Enum]) - The current direction Karel is facing.
        """
        return self._direction

    @direction.setter
    def direction(self, val):
        """
        This property can be used to set Karel's current direction.

        Parameters:
            val (Direction[Enum]) - The new direction that Karel will be facing.
        Returns: None
        """
        self._direction = val

    @property
    def num_beepers(self):
        """
        This property can be used to access Karel's current number of beepers.

        Parameters: None
        Returns:
            num_beepers (int) - The current number of beepers Karel has.
        """
        return self._num_beepers

    @num_beepers.setter
    def num_beepers(self, val):
        """
        This property can be used to set Karel's current number of beepers.

        Parameters:
            val (int) - The new number of beepers that Karel will have.
        Returns: None
        """
        self._num_beepers = val

    def reset_state(self):
        """
        This function is used to reset Karel's location and direction to the original
        starting parameters as indicated by the world that Karel lives in.

        Parameters: None
        Returns: None
        """
        self._avenue, self._street = self.world.karel_starting_location
        self._direction = self.world.karel_starting_direction

        self._num_beepers = self.world.karel_starting_beeper_count

    def move(self):
        """
        This function moves Karel forward one space in the direction that it is
        currently facing. If Karel's front is not clear (blocked by wall or boundary
        of world) then a KarelException will be raised).

        Parameters: None
        Returns: None
        """
        if not self.front_is_clear():
            raise KarelException(
                self._avenue,
                self._street,
                self._direction,
                "Karel attempted to move, but its front was blocked.",
            )

        delta_avenue, delta_street = DIRECTION_DELTA_MAP[self._direction]
        self._avenue += delta_avenue
        self._street += delta_street

    def turn_left(self):
        """
        This function turns Karel 90 degrees counterclockwise.

        Parameters: None
        Returns: None
        """
        self._direction = NEXT_DIRECTION_MAP[self._direction]

    def put_beeper(self):
        """
        This function places a beeper on the corner that Karel is currently standing
        on and decreases Karel's beeper count by 1. If Karel has no more beepers in its
        beeper bag, then this function raises a KarelException.

        Parameters: None
        Returns: None
        """
        if self._num_beepers == 0:
            raise KarelException(
                self._avenue,
                self._street,
                self._direction,
                "Karel attempted to put a beeper, but it had none left in its bag.",
            )

        if self._num_beepers != INFINITY:
            self._num_beepers -= 1

        self.world.add_beeper(self._avenue, self._street)

    def pick_beeper(self):
        """
        This function removes a beeper from the corner that Karel is currently standing on
        and increases Karel's beeper count by 1. If there are no beepers on Karel's current
        corner, then this function raises a KarelException.

        Parameters: None
        Returns: None
        """
        if not self.beepers_present():
            raise KarelException(
                self._avenue,
                self._street,
                self._direction,
                "Karel attempted to pick up a beeper, but there were none on the current corner.",
            )

        if self._num_beepers != INFINITY:
            self._num_beepers += 1

        self.world.remove_beeper(self._avenue, self._street)

    def front_is_clear(self):
        """
        This function returns a boolean indicating whether or not there is a wall
        in front of Karel.

        Parameters: None
        Returns:
            is_clear (Bool) - True if there is no wall in front of Karel
                              False otherwise
        """
        return self.direction_is_clear(self._direction)

    def direction_is_clear(self, direction):
        """
        This is a helper function that returns a boolean indicating whether
        or not there is a barrier in the specified direction of Karel.

        Parameters:
            direction (Direction[Enum]) - The direction in which to check for a barrier

        Returns:
            is_clear (Bool) - True if there is no barrier in the specified direction
                              False otherwise
        """
        delta_avenue, delta_street = DIRECTION_DELTA_MAP[direction]
        next_avenue = self._avenue + delta_avenue
        next_street = self._street + delta_street

        # front is not clear if we are about to go out of bounds
        if not self.world.in_bounds(next_avenue, next_street):
            return False

        # front is not clear if wall exists in same direction of where we're currently facing
        if self.world.wall_exists(self._avenue, self._street, direction):
            return False

        # must also check for alternate possible representation of wall
        opposite_direction = NEXT_DIRECTION_MAP[NEXT_DIRECTION_MAP[direction]]
        if self.world.wall_exists(next_avenue, next_street, opposite_direction):
            return False

        # If all previous conditions checked out, then the front is clear
        return True

    def front_is_blocked(self):
        """
        This function returns a boolean indicating whether there is a wall
        in front of Karel.

        Parameters: None
        Returns:
            is_blocked (Bool) - True if there is a wall in front of Karel
                                  False otherwise
        """
        return not self.front_is_clear()

    def left_is_clear(self):
        """
        This function returns a boolean indicating whether or not there is a wall
        to the left of Karel.

        Parameters: None
        Returns:
            is_clear (Bool) - True if there is no wall to the left of Karel
                              False otherwise
        """
        return self.direction_is_clear(NEXT_DIRECTION_MAP[self._direction])

    def left_is_blocked(self):
        """
        This function returns a boolean indicating whether there is a wall
        to the left of Karel.

        Parameters: None
        Returns:
            is_blocked (Bool) - True if there is a wall to the left of Karel
                                  False otherwise
        """
        return not self.left_is_clear()

    def right_is_clear(self):
        """
        This function returns a boolean indicating whether or not there is a wall
        to the right of Karel.

        Parameters: None
        Returns:
            is_clear (Bool) - True if there is no wall to the right of Karel
                              False otherwise
        """
        return self.direction_is_clear(NEXT_DIRECTION_MAP_RIGHT[self._direction])

    def right_is_blocked(self):
        """
        This function returns a boolean indicating whether there is a wall
        to the right of Karel.

        Parameters: None
        Returns:
            is_blocked (Bool) - True if there is a wall to the right of Karel
                                  False otherwise
        """
        return not self.right_is_clear()

    def beepers_present(self):
        """
        This function returns a boolean indicating whether or not there is
        a beeper on Karel's current corner.

        Parameters: None
        Returns:
            beepers_on_corner (Bool) - True if there's at least one beeper on Karel's current corner
                                       False otherwise
        """
        return self.world.beepers[(self.avenue, self.street)] != 0

    def no_beepers_present(self):
        return not self.beepers_present()

    def beepers_in_bag(self):
        """
        This function returns a boolean indicating whether or not there is
        at least one beeper in Karel's beeper bag.

        Parameters: None
        Returns:
            beepers_in_bag (Bool) - True if there is at least one beeper in Karel's bag
                                    False otherwise
        """
        # Can't check > 0 because INFINITY beepers is -1
        return self._num_beepers != 0

    def no_beepers_in_bag(self):
        # Only 0 beepers in bag indicates empty bag – negative numbers represent INFINITY
        return self._num_beepers == 0

    def facing_north(self):
        """
        This function returns a boolean indicating whether or not Karel is currently
        facing North.

        Parameters: None
        Returns:
            facing_north (Bool) - True if Karel is currently facing North
                                  False otherwise
        """
        return self.direction == Direction.NORTH

    def not_facing_north(self):
        return not self.facing_north()

    def facing_east(self):
        """
        This function returns a boolean indicating whether or not Karel is currently
        facing East.

        Parameters: None
        Returns:
            facing_east (Bool) - True if Karel is currently facing East
                                 False otherwise
        """
        return self.direction == Direction.EAST

    def not_facing_east(self):
        return not self.facing_east()

    def facing_west(self):
        """
        This function returns a boolean indicating whether or not Karel is currently
        facing West.

        Parameters: None
        Returns:
            facing_west (Bool) - True if Karel is currently facing West
                                 False otherwise
        """
        return self.direction == Direction.WEST

    def not_facing_west(self):
        return not self.facing_west()

    def facing_south(self):
        """
        This function returns a boolean indicating whether or not Karel is currently
        facing South.

        Parameters: None
        Returns:
            facing_south (Bool) - True if Karel is currently facing South
                                  False otherwise
        """
        return self.direction == Direction.SOUTH

    def not_facing_south(self):
        return not self.facing_south()

    def paint_corner(self, color):
        """
        This function makes Karel paint its current corner the indicated color.
        This function will raise a KarelException if the indicated color is not one
        of the valid predefined colors. For this list of colors, check the
        kareldefinitions.py file.

        Parameters:
            color (str) - The color string specifying which color to paint the corner
        Returns: None
        """
        if color is not None and color not in COLOR_MAP:
            raise KarelException(
                self._avenue,
                self._street,
                self._direction,
                "Karel attempted to paint the corner with color {}, which is not valid.".format(
                    color
                ),
            )
        self.world.paint_corner(self.avenue, self.street, color)

    def corner_color_is(self, color):
        """
        This function returns a boolean indicating whether or not Karel's current
        corner is the specified color.

        Parameters:
            color (str) - Color string representing the color to check the current corner for
        Returns:
            is_color (Bool) - True if Karel's current corner is the specified color
                              False otherwise
        """
        return self.world.corner_color(self.avenue, self.street) == color



"""
This file contains all of the constant and class definitions used
throughout the Karel library. Constants here define drawing proportions
for Karel and its world, initial world parameters, and different exception
and enumeration types, among other things.

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""
import math
from enum import Enum, unique


@unique
class Direction(Enum):
    EAST = 0
    SOUTH = math.pi / 2
    WEST = math.pi
    NORTH = 3 * math.pi / 2


# Karel Application + World Editor
PAD_X = 75
PAD_Y = 10
LIGHT_GREY = "#e5e5e5"

# Karel Application + World Editor + Karel Canvas
DEFAULT_ICON = "karel"

# Karel, World Editor, Karel World
INFINITY = -1
COLOR_MAP = {
    "Red": "red",
    "Black": "black",
    "Cyan": "cyan",
    "Dark Gray": "gray30",
    "Gray": "gray55",
    "Green": "green",
    "Light Gray": "gray80",
    "Magenta": "magenta3",
    "Orange": "orange",
    "Pink": "pink",
    "White": "snow",
    "Blue": "blue",
    "Yellow": "yellow",
}

# Karel World, World Editor
DIRECTIONS_MAP = {
    "north": Direction.NORTH,
    "east": Direction.EAST,
    "south": Direction.SOUTH,
    "west": Direction.WEST,
}
DIRECTIONS_MAP_INVERSE = {v: k for k, v in DIRECTIONS_MAP.items()}


# Karel World, Karel Canvas
class Wall:
    """ Note that the World Editor only uses West & South to denote wall directions. """

    def __init__(self, avenue, street, direction):
        self._avenue = avenue
        self._street = street
        self._direction = direction

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash((self._avenue, self._street, self._direction))

    def __repr__(self):
        return "({}, {}) {}".format(self._avenue, self._street, self._direction)

    @property
    def avenue(self):
        return self._avenue

    @property
    def street(self):
        return self._street

    @property
    def direction(self):
        return self._direction


# Karel Application + Karel
class KarelException(Exception):
    """
	The following classes define Karel-specific exceptions.
	"""

    def __init__(self, avenue, street, direction, message):
        super().__init__()
        self.avenue = avenue
        self.street = street
        self.direction = DIRECTIONS_MAP_INVERSE[direction].capitalize()
        self.message = message

    def __str__(self):
        return (
            "Karel crashed while on avenue {} and street {}, "
            "facing {}\nInvalid action: {}".format(
                self.avenue, self.street, self.direction, self.message
            )
        )



"""
This file defines the canvas upon which a Karel world is drawn. This
class defines all necessary methods to draw all components of a Karel
world, including walls, beepers, and Karel itself. All Karel applications
contains exactly one Karel Canvas object and each Karel Canvas object
holds information about one Karel World and one Karel object.

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""

import cmath
import math
import tkinter as tk

from .karel_definitions import DEFAULT_ICON, Direction, Wall

WALL_DETECTION_THRESHOLD = 0.1
BORDER_OFFSET = 17
LABEL_OFFSET = 7
CORNER_SIZE = 2
BEEPER_CELL_SIZE_FRAC = 0.4
LINE_WIDTH = 2
# Drawing Constants for Karel Robot Icon (defined relative to a single cell)
KAREL_VERTICAL_OFFSET = 0.05
KAREL_LEFT_HORIZONTAL_PAD = 0.29
KAREL_HEIGHT = 0.76
KAREL_WIDTH = 0.58
KAREL_INNER_HEIGHT = 0.38
KAREL_INNER_WIDTH = 0.28125
KAREL_INNER_OFFSET = 0.125
KAREL_MOUTH_WIDTH = 0.1375
KAREL_MOUTH_HORIZONTAL_OFFSET = 0.2625
KAREL_MOUTH_VERTICAL_OFFSET = 0.125
KAREL_UPPER_RIGHT_DIAG = 0.2
KAREL_LOWER_LEFT_DIAG = 0.13125
KAREL_LEG_LENGTH = 0.15
KAREL_FOOT_LENGTH = 0.1875
KAREL_LEG_FOOT_WIDTH = 0.075
KAREL_LEG_VERTICAL_OFFSET = 0.5
KAREL_LEG_HORIZONTAL_OFFSET = 0.2625
KAREL_LINE_WIDTH = 2
# Drawing Constants for Simple Karel Icon (defined relative to a single cell)
SIMPLE_KAREL_HEIGHT = 0.7
SIMPLE_KAREL_WIDTH = 0.8


class KarelCanvas(tk.Canvas):
    def __init__(self, width, height, master, world=None, karel=None, bg="white"):
        super().__init__(master, width=width, height=height, bg=bg)
        self.world = world
        self.karel = karel
        self.icon = DEFAULT_ICON
        self.draw_world()
        self.draw_karel()

    def create_polygon(
        self, points, fill="black", outline="black", width=KAREL_LINE_WIDTH, tag="karel"
    ):
        super().create_polygon(points, fill=fill, outline=outline, width=width, tag=tag)

    def set_icon(self, icon):
        self.icon = icon

    def redraw_all(self):
        self.delete("all")
        self.draw_world()
        self.draw_karel()
        self.update()

    def redraw_karel(self, update=True):
        self.delete("karel")
        self.draw_karel()
        if update:
            self.update()

    def redraw_beepers(self, update=True):
        self.delete("beeper")
        self.draw_all_beepers()
        if update:
            self.update()

    def redraw_corners(self, update=True):
        self.delete("corner")
        self.draw_corners()
        if update:
            self.update()

    def redraw_walls(self, update=True):
        self.delete("wall")
        self.draw_all_walls()
        if update:
            self.update()

    def draw_world(self):
        self.init_geometry_values()
        self.draw_bounding_rectangle()
        self.label_axes()
        self.draw_corners()
        self.draw_all_beepers()
        self.draw_all_walls()

    def init_geometry_values(self):
        self.update()

        # Calculate the maximum possible cell size in both directions
        # We will use the smaller of the two as the bounding cell size
        horizontal_cell_size = (self.winfo_width() - 2 * BORDER_OFFSET) / self.world.num_avenues
        vertical_cell_size = (self.winfo_height() - 2 * BORDER_OFFSET) / self.world.num_streets

        # Save this as an instance variable for later use
        self.cell_size = min(horizontal_cell_size, vertical_cell_size)

        self.boundary_height = self.cell_size * self.world.num_streets
        self.boundary_width = self.cell_size * self.world.num_avenues

        # Save all these as instance variables as well
        self.left_x = self.winfo_width() / 2 - self.boundary_width / 2
        self.top_y = self.winfo_height() / 2 - self.boundary_height / 2
        self.right_x = self.left_x + self.boundary_width
        self.bottom_y = self.top_y + self.boundary_height

    def draw_bounding_rectangle(self):
        # Draw the external bounding lines of Karel's world
        self.create_line(self.left_x, self.top_y, self.right_x, self.top_y, width=LINE_WIDTH)
        self.create_line(self.left_x, self.top_y, self.left_x, self.bottom_y, width=LINE_WIDTH)
        self.create_line(self.right_x, self.top_y, self.right_x, self.bottom_y, width=LINE_WIDTH)
        self.create_line(self.left_x, self.bottom_y, self.right_x, self.bottom_y, width=LINE_WIDTH)

    def label_axes(self):
        # Label the avenue axes
        for avenue in range(1, self.world.num_avenues + 1):
            label_x = self.calculate_corner_x(avenue)
            label_y = self.bottom_y + LABEL_OFFSET
            self.create_text(label_x, label_y, text=str(avenue), font="Arial 10")

        # Label the street axes
        for street in range(1, self.world.num_streets + 1):
            label_x = self.left_x - LABEL_OFFSET
            label_y = self.calculate_corner_y(street)
            self.create_text(label_x, label_y, text=str(street), font="Arial 10")

    def draw_corners(self):
        # Draw all corner markers in the world
        for avenue in range(1, self.world.num_avenues + 1):
            for street in range(1, self.world.num_streets + 1):
                color = self.world.corner_color(avenue, street)
                corner_x = self.calculate_corner_x(avenue)
                corner_y = self.calculate_corner_y(street)
                if not color:
                    self.create_line(
                        corner_x,
                        corner_y - CORNER_SIZE,
                        corner_x,
                        corner_y + CORNER_SIZE,
                        tag="corner",
                    )
                    self.create_line(
                        corner_x - CORNER_SIZE,
                        corner_y,
                        corner_x + CORNER_SIZE,
                        corner_y,
                        tag="corner",
                    )
                else:
                    self.create_rectangle(
                        corner_x - self.cell_size / 2,
                        corner_y - self.cell_size / 2,
                        corner_x + self.cell_size / 2,
                        corner_y + self.cell_size / 2,
                        fill=color,
                        tag="corner",
                        outline="",
                    )

    def draw_all_beepers(self):
        for location, count in self.world.beepers.items():
            self.draw_beeper(location, count)

    def draw_beeper(self, location, count):
        # handle case where defaultdict returns 0 count by not drawing beepers
        if count == 0:
            return

        corner_x = self.calculate_corner_x(location[0])
        corner_y = self.calculate_corner_y(location[1])
        beeper_radius = self.cell_size * BEEPER_CELL_SIZE_FRAC

        points = [
            corner_x,
            corner_y - beeper_radius,
            corner_x + beeper_radius,
            corner_y,
            corner_x,
            corner_y + beeper_radius,
            corner_x - beeper_radius,
            corner_y,
        ]
        self.create_polygon(points, fill="light grey", tag="beeper")

        if count > 1:
            self.create_text(corner_x, corner_y, text=str(count), font="Arial 12", tag="beeper")

    def draw_all_walls(self):
        for wall in self.world.walls:
            self.draw_wall(wall)

    def draw_wall(self, wall):
        avenue, street, direction = wall.avenue, wall.street, wall.direction
        corner_x = self.calculate_corner_x(avenue)
        corner_y = self.calculate_corner_y(street)

        if direction == Direction.NORTH:
            self.create_line(
                corner_x - self.cell_size / 2,
                corner_y - self.cell_size / 2,
                corner_x + self.cell_size / 2,
                corner_y - self.cell_size / 2,
                width=LINE_WIDTH,
                tag="wall",
            )
        if direction == Direction.SOUTH:
            self.create_line(
                corner_x - self.cell_size / 2,
                corner_y + self.cell_size / 2,
                corner_x + self.cell_size / 2,
                corner_y + self.cell_size / 2,
                width=LINE_WIDTH,
                tag="wall",
            )
        if direction == Direction.EAST:
            self.create_line(
                corner_x + self.cell_size / 2,
                corner_y - self.cell_size / 2,
                corner_x + self.cell_size / 2,
                corner_y + self.cell_size / 2,
                width=LINE_WIDTH,
                tag="wall",
            )
        if direction == Direction.WEST:
            self.create_line(
                corner_x - self.cell_size / 2,
                corner_y - self.cell_size / 2,
                corner_x - self.cell_size / 2,
                corner_y + self.cell_size / 2,
                width=LINE_WIDTH,
                tag="wall",
            )

    def draw_karel(self):
        corner_x = self.calculate_corner_x(self.karel.avenue)
        corner_y = self.calculate_corner_y(self.karel.street)
        center = (corner_x, corner_y)

        if self.icon == "karel":
            karel_origin_x = (
                corner_x - self.cell_size / 2 + KAREL_LEFT_HORIZONTAL_PAD * self.cell_size
            )
            karel_origin_y = corner_y - self.cell_size / 2 + KAREL_VERTICAL_OFFSET * self.cell_size

            self.draw_karel_body(karel_origin_x, karel_origin_y, center, self.karel.direction.value)
            self.draw_karel_legs(karel_origin_x, karel_origin_y, center, self.karel.direction.value)
        elif self.icon == "simple":
            self.draw_simple_karel_icon(center, self.karel.direction.value)

    def generate_external_karel_points(self, x, y, center, direction):
        outer_points = []

        # Top-left point (referred to as origin) of Karel's body
        outer_points += [x, y]

        # Calculate Karel's height and width as well as missing diag segments
        width = self.cell_size * KAREL_WIDTH
        height = self.cell_size * KAREL_HEIGHT
        lower_left_missing = (self.cell_size * KAREL_LOWER_LEFT_DIAG) / math.sqrt(2)
        upper_right_missing = (self.cell_size * KAREL_UPPER_RIGHT_DIAG) / math.sqrt(2)

        # These two points define Karel's upper right
        outer_points += [x + width - upper_right_missing, y]
        outer_points += [x + width, y + upper_right_missing]

        # Karel's bottom right edge
        outer_points += [x + width, y + height]

        # These two points define Karel's lower left
        outer_points += [x + lower_left_missing, y + height]
        outer_points += [x, y + height - lower_left_missing]

        # Complete the polygon
        outer_points += [x, y]

        # Rotate all external body points to get correct Karel orientation
        self.rotate_points(center, outer_points, direction)

        return outer_points

    def generate_internal_karel_points(self, x, y, center, direction):

        # Calculate dimensions and location of Karel's inner eye
        inner_x = x + self.cell_size * KAREL_INNER_OFFSET
        inner_y = y + self.cell_size * KAREL_INNER_OFFSET
        inner_height = self.cell_size * KAREL_INNER_HEIGHT
        inner_width = self.cell_size * KAREL_INNER_WIDTH

        # Define inner body points
        inner_points = [
            inner_x,
            inner_y,
            inner_x + inner_width,
            inner_y,
            inner_x + inner_width,
            inner_y + inner_height,
            inner_x,
            inner_y + inner_height,
            inner_x,
            inner_y,
        ]
        self.rotate_points(center, inner_points, direction)

        return inner_points

    def draw_karel_body(self, x, y, center, direction):
        outer_points = self.generate_external_karel_points(x, y, center, direction)
        inner_points = self.generate_internal_karel_points(x, y, center, direction)

        # Non-convex polygon that determines Karel's entire body is a
        # combination of the two sets of points defining internal and external components
        entire_body_points = outer_points + inner_points

        # First draw the filled non-convex polygon
        self.create_polygon(entire_body_points, fill="white", outline="")

        # Then draw the transparent exterior edges of Karel's body
        self.create_polygon(outer_points, fill="")
        self.create_polygon(inner_points, fill="")

        # Define dimensions and location of Karel's mouth
        # karel_height = self.cell_size * KAREL_HEIGHT
        mouth_horizontal_offset = self.cell_size * KAREL_MOUTH_HORIZONTAL_OFFSET
        mouth_vertical_offset = self.cell_size * KAREL_MOUTH_VERTICAL_OFFSET
        inner_y = y + self.cell_size * KAREL_INNER_OFFSET
        inner_height = self.cell_size * KAREL_INNER_HEIGHT
        mouth_width = self.cell_size * KAREL_MOUTH_WIDTH

        mouth_y = inner_y + inner_height + mouth_vertical_offset

        # Define, rotate, and draw points
        mouth_points = [
            x + mouth_horizontal_offset,
            mouth_y,
            x + mouth_horizontal_offset + mouth_width,
            mouth_y,
        ]
        self.rotate_points(center, mouth_points, direction)
        self.create_polygon(mouth_points, fill="white")

    def draw_karel_legs(self, x, y, center, direction):
        leg_length = self.cell_size * KAREL_LEG_LENGTH
        foot_length = self.cell_size * KAREL_FOOT_LENGTH
        leg_foot_width = self.cell_size * KAREL_LEG_FOOT_WIDTH

        vertical_offset = self.cell_size * KAREL_LEG_VERTICAL_OFFSET
        horizontal_offset = self.cell_size * KAREL_LEG_HORIZONTAL_OFFSET

        # Generate points for left leg
        points = []
        points += [x, y + vertical_offset]
        points += [x - leg_length, y + vertical_offset]
        points += [x - leg_length, y + vertical_offset + foot_length]
        points += [x - leg_length + leg_foot_width, y + vertical_offset + foot_length]
        points += [x - leg_length + leg_foot_width, y + vertical_offset + leg_foot_width]
        points += [x, y + vertical_offset + leg_foot_width]
        points += [x, y + vertical_offset]

        self.rotate_points(center, points, direction)
        self.create_polygon(points)

        # Reset point of reference to be bottom left rather than top_left
        y += self.cell_size * KAREL_HEIGHT

        # Generate points for right leg
        points = []
        points += [x + horizontal_offset, y]
        points += [x + horizontal_offset, y + leg_length]
        points += [x + horizontal_offset + foot_length, y + leg_length]
        points += [x + horizontal_offset + foot_length, y + leg_length - leg_foot_width]
        points += [x + horizontal_offset + leg_foot_width, y + leg_length - leg_foot_width]
        points += [x + horizontal_offset + leg_foot_width, y]
        points += [x + horizontal_offset, y]

        self.rotate_points(center, points, direction)
        self.create_polygon(points)

    def draw_simple_karel_icon(self, center, direction):
        simple_karel_width = self.cell_size * SIMPLE_KAREL_WIDTH
        simple_karel_height = self.cell_size * SIMPLE_KAREL_HEIGHT
        center_x, center_y = center
        points = []
        points += [center_x - simple_karel_width / 2, center_y - simple_karel_height / 2]
        points += [center_x - simple_karel_width / 2, center_y + simple_karel_height / 2]
        points += [center_x, center_y + simple_karel_height / 2]
        points += [center_x + simple_karel_width / 2, center_y]
        points += [center_x, center_y - simple_karel_height / 2]
        points += [center_x - simple_karel_width / 2, center_y - simple_karel_height / 2]
        self.rotate_points(center, points, direction)
        self.create_polygon(points, fill="white")

    def calculate_corner_x(self, avenue):
        return self.left_x + self.cell_size / 2 + (avenue - 1) * self.cell_size

    def calculate_corner_y(self, street):
        return self.top_y + self.cell_size / 2 + (self.world.num_streets - street) * self.cell_size

    def click_in_world(self, x, y):
        x = x - self.left_x
        y = y - self.top_y
        return 0 <= x < self.boundary_width and 0 <= y < self.boundary_height

    def calculate_location(self, x, y):
        x = x - self.left_x
        y = y - self.top_y
        return (
            int(max(x, 0) // self.cell_size) + 1,
            int(max((self.boundary_height - 1 - y), 0) // self.cell_size) + 1,
        )

    def find_nearest_wall(self, x, y, avenue, street):
        corner_x = self.calculate_corner_x(avenue)
        corner_y = self.calculate_corner_y(street)
        wall_proximity = self.cell_size * WALL_DETECTION_THRESHOLD

        if x > (corner_x + self.cell_size / 2 - wall_proximity):
            # Check for a wall to the east
            return Wall(avenue, street, Direction.EAST)
        if x < (corner_x - self.cell_size / 2 + wall_proximity):
            # Check for a wall to the west
            return Wall(avenue, street, Direction.WEST)
        if y > (corner_y + self.cell_size / 2 - wall_proximity):
            # Check for a wall to the south
            return Wall(avenue, street, Direction.SOUTH)
        if y < (corner_y - self.cell_size / 2 + wall_proximity):
            # Check for a wall to the north
            return Wall(avenue, street, Direction.NORTH)

        # No wall within threshold distance
        return None

    @staticmethod
    def rotate_points(center, points, direction):
        """
		Rotation logic derived from http://effbot.org/zone/tkinter-complex-canvas.htm
		"""
        cangle = cmath.exp(direction * 1j)
        center = complex(center[0], center[1])
        for i in range(0, len(points), 2):
            x, y = points[i], points[i + 1]
            v = cangle * (complex(x, y) - center) + center
            points[i], points[i + 1] = v.real, v.imag



"""
This file defines the TK GUI for editing Karel worlds.

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""

import os
import tkinter as tk
from tkinter import messagebox, simpledialog
from tkinter.filedialog import askopenfilename, asksaveasfilename

from .karel import Karel
from .karel_canvas import KarelCanvas
from .karel_definitions import (
    COLOR_MAP,
    DEFAULT_ICON,
    DIRECTIONS_MAP,
    DIRECTIONS_MAP_INVERSE,
    INFINITY,
    LIGHT_GREY,
    PAD_X,
    PAD_Y,
)

MIN_DIMENSIONS = 1
MAX_DIMENSIONS = 50
DEFAULT_COLOR = "Red"
DEFAULT_SIZE = 8


def run_world_editor():
    root = tk.Tk()
    world_builder = WorldBuilderApplication(master=root)
    world_builder.mainloop()


class WorldBuilderApplication(tk.Frame):
    def __init__(
        self, master=None, window_width=800, window_height=400, canvas_width=600, canvas_height=400
    ):
        # set window background to contrast white Karel canvas
        master.configure(background=LIGHT_GREY)

        # configure location of canvas to expand to fit window resizing
        master.rowconfigure(0, weight=1)
        master.columnconfigure(1, weight=1)

        super().__init__(master, width=window_width, height=window_height, background=LIGHT_GREY)
        self.icon = DEFAULT_ICON
        self.window_width = window_width
        self.window_height = window_height
        self.canvas_width = canvas_width
        self.canvas_height = canvas_height
        self.last_action_event_loc = (None, None)
        self.master = master
        self.master.title("Karel World Builder")
        self.set_dock_icon()
        self.grid(row=0, column=0)
        self.master.update()
        self.setup_world()
        self.create_canvas()
        self.create_buttons()

    def set_dock_icon(self):
        # make Karel dock icon image
        path = os.path.join(os.path.dirname(__file__), "icon.png")
        img = tk.Image("photo", file=path)
        self.master.tk.call("wm", "iconphoto", self.master._w, img)

    def setup_world(self):
        load_existing = messagebox.askyesno(
            "World Selection",
            "Would you like to load an existing world? \n\n"
            "Selecting 'No' will allow you to start with a blank slate.",
            parent=self.master,
        )
        if load_existing:
            self.load_world(init=True)
        else:
            self.create_new_world(init=True, default=True)

    def create_new_world(self, init=False, default=False):
        num_avenues = simpledialog.askinteger(
            "New World Size",
            "How many avenues should the new world have?",
            parent=self.master,
            minvalue=MIN_DIMENSIONS,
            maxvalue=MAX_DIMENSIONS,
        )

        if not num_avenues:
            if default:
                num_avenues = DEFAULT_SIZE
            else:
                # In this case, we can just cancel execution and return to existing world
                return

        num_streets = simpledialog.askinteger(
            "New World Size",
            "How many streets should the new world have?",
            parent=self.master,
            minvalue=MIN_DIMENSIONS,
            maxvalue=MAX_DIMENSIONS,
        )
        if not num_streets:
            if default:
                num_streets = DEFAULT_SIZE
            else:
                # In this case, we can just cancel execution and return to existing world
                return
        if init:
            self.karel = Karel()
            self.world = self.karel.world
        else:
            self.world.reload_world()
            self.karel.reset_state()

        self.world.num_avenues = num_avenues
        self.world.num_streets = num_streets
        if not init:
            self.canvas.redraw_all()

    def load_world(self, init=False):
        default_worlds_path = os.path.join(os.path.dirname(__file__), "worlds")
        filename = askopenfilename(
            initialdir=default_worlds_path,
            title="Select Karel World",
            filetypes=[("Karel Worlds", "*.w")],
            parent=self.master,
        )
        # User hit cancel and did not select file, so leave world as-is
        if filename == "":
            if init:
                self.setup_world()
            return

        if init:
            self.karel = Karel(filename)
            self.world = self.karel.world
        else:
            self.world.reload_world(filename=filename)
            self.karel.reset_state()
            self.canvas.redraw_all()
            self.reset_direction_radio_buttons()
            self.reset_beeper_bag_radio_buttons()

    def create_canvas(self):
        """
        This method creates the canvas on which Karel and Karel's
        world are drawn.
        """
        self.canvas = KarelCanvas(
            self.canvas_width, self.canvas_height, self.master, world=self.world, karel=self.karel
        )
        self.canvas.grid(column=1, row=0, sticky="NESW")
        self.canvas.bind("<Configure>", lambda t: self.canvas.redraw_all())
        self.canvas.bind("<Button-1>", self.handle_mouse_event)
        self.canvas.bind("<B1-Motion>", self.handle_mouse_event)

    def create_buttons(self):
        """
        This method creates the three buttons that appear on the left
        side of the screen. These buttons control the start of Karel
        execution, resetting Karel's state, and loading new worlds.
        """
        self.program_control_button = tk.Button(
            self, highlightthickness=0, highlightbackground="white"
        )
        self.program_control_button["text"] = "New World"
        self.program_control_button["command"] = self.create_new_world
        self.program_control_button.grid(column=0, row=0, padx=PAD_X, pady=PAD_Y)

        self.load_world_button = tk.Button(
            self, highlightthickness=0, text="Load World", command=self.load_world
        )
        self.load_world_button.grid(column=0, row=1, padx=PAD_X, pady=PAD_Y)

        self.save_world_button = tk.Button(
            self, highlightthickness=0, text="Save World", command=self.save_world
        )
        self.save_world_button.grid(column=0, row=2, padx=PAD_X, pady=PAD_Y)

        self.create_direction_radio_buttons()
        self.create_beeper_bag_radio_buttons()
        self.create_action_radio_buttons()

    def create_direction_radio_buttons(self):
        self.dir_radio_frame = tk.Frame(self, bg=LIGHT_GREY)
        self.dir_radio_frame.grid(row=3, column=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.karel_direction_var = tk.StringVar()
        self.karel_direction_var.set(DIRECTIONS_MAP_INVERSE[self.karel.direction])
        self.karel_direction_var.trace("w", self.update_karel_direction)

        dir_label = tk.Label(self.dir_radio_frame, text="Karel Direction: ", bg=LIGHT_GREY)
        dir_label.pack(side="left")
        tk.Radiobutton(
            self.dir_radio_frame,
            text="E",
            variable=self.karel_direction_var,
            value="east",
            bg=LIGHT_GREY,
        ).pack(side="left")
        tk.Radiobutton(
            self.dir_radio_frame,
            text="W",
            variable=self.karel_direction_var,
            value="west",
            bg=LIGHT_GREY,
        ).pack(side="left")
        tk.Radiobutton(
            self.dir_radio_frame,
            text="N",
            variable=self.karel_direction_var,
            value="north",
            bg=LIGHT_GREY,
        ).pack(side="left")
        tk.Radiobutton(
            self.dir_radio_frame,
            text="S",
            variable=self.karel_direction_var,
            value="south",
            bg=LIGHT_GREY,
        ).pack(side="left")

    def create_beeper_bag_radio_buttons(self):
        self.beeper_bag_radio_frame = tk.Frame(self, bg=LIGHT_GREY)
        self.beeper_bag_radio_frame.grid(row=4, column=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.beeper_bag_var = tk.IntVar()
        self.beeper_bag_var.set(self.karel.num_beepers)
        self.beeper_bag_var.trace("w", self.update_karel_num_beepers)

        beeper_bag_label = tk.Label(self.beeper_bag_radio_frame, text="Beeper Bag: ", bg=LIGHT_GREY)
        beeper_bag_label.pack(side="left")
        tk.Radiobutton(
            self.beeper_bag_radio_frame,
            text="Empty",
            variable=self.beeper_bag_var,
            value=0,
            bg=LIGHT_GREY,
        ).pack(side="left")
        tk.Radiobutton(
            self.beeper_bag_radio_frame,
            text="Infinite",
            variable=self.beeper_bag_var,
            value=INFINITY,
            bg=LIGHT_GREY,
        ).pack(side="left")

    def create_action_radio_buttons(self):
        self.action_radio_frame = tk.Frame(self, bg=LIGHT_GREY)
        self.action_radio_frame.grid(row=5, column=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.action_var = tk.StringVar()
        self.action_var.set("move_karel")

        action_label = tk.Label(self.action_radio_frame, text="Action: ", bg=LIGHT_GREY)
        action_label.pack(side="left")
        tk.Radiobutton(
            self.action_radio_frame,
            text="Move Karel",
            variable=self.action_var,
            value="move_karel",
            bg=LIGHT_GREY,
        ).pack(anchor="w")
        tk.Radiobutton(
            self.action_radio_frame,
            text="Add Wall",
            variable=self.action_var,
            value="add_wall",
            bg=LIGHT_GREY,
        ).pack(anchor="w")
        tk.Radiobutton(
            self.action_radio_frame,
            text="Remove Wall",
            variable=self.action_var,
            value="remove_wall",
            bg=LIGHT_GREY,
        ).pack(anchor="w")
        tk.Radiobutton(
            self.action_radio_frame,
            text="Add Beeper",
            variable=self.action_var,
            value="add_beeper",
            bg=LIGHT_GREY,
        ).pack(anchor="w")
        tk.Radiobutton(
            self.action_radio_frame,
            text="Remove Beeper",
            variable=self.action_var,
            value="remove_beeper",
            bg=LIGHT_GREY,
        ).pack(anchor="w")

        color_selection_frame = tk.Frame(self.action_radio_frame, bg=LIGHT_GREY)
        color_selection_frame.pack(anchor="w")

        self.color_var = tk.StringVar()
        self.color_var.set(DEFAULT_COLOR)

        tk.Radiobutton(
            color_selection_frame,
            text="Paint Corner",
            variable=self.action_var,
            value="paint_corner",
            bg=LIGHT_GREY,
        ).pack(side="left")
        self.color_dropdown = tk.OptionMenu(
            color_selection_frame, self.color_var, *sorted(COLOR_MAP.keys())
        )
        self.color_dropdown["bg"] = LIGHT_GREY
        self.color_dropdown.pack(side="left")

        tk.Radiobutton(
            self.action_radio_frame,
            text="Reset Corner",
            variable=self.action_var,
            value="reset_corner",
            bg=LIGHT_GREY,
        ).pack(anchor="w")

    def reset_direction_radio_buttons(self):
        self.karel_direction_var.set(DIRECTIONS_MAP_INVERSE[self.karel.direction])

    def reset_beeper_bag_radio_buttons(self):
        self.beeper_bag_var.set(self.karel.num_beepers)

    def update_karel_direction(self, *args):
        del args
        new_dir = self.karel_direction_var.get()
        self.karel.direction = DIRECTIONS_MAP[new_dir]
        self.canvas.redraw_karel()

    def update_karel_num_beepers(self, *args):
        del args
        new_num_beepers = self.beeper_bag_var.get()
        self.karel.num_beepers = new_num_beepers

    def handle_mouse_event(self, event):
        def apply_function(fn, *args):
            if event_type == tk.EventType.ButtonPress:
                self.last_action_event_loc = (avenue, street)
                fn(avenue, street, *args)
                self.canvas.redraw_corners(update=False)
                self.canvas.redraw_beepers(update=False)
                self.canvas.redraw_walls(update=False)
                self.canvas.redraw_karel()

            elif event_type == tk.EventType.Motion:
                if (avenue, street) != self.last_action_event_loc:
                    self.last_action_event_loc = (avenue, street)
                    fn(avenue, street, *args)
                    self.canvas.redraw_corners(update=False)
                    self.canvas.redraw_beepers(update=False)
                    self.canvas.redraw_walls(update=False)
                    self.canvas.redraw_karel()

        event_type = event.type
        # only handle click events that happen in the world
        if not self.canvas.click_in_world(event.x, event.y):
            return

        avenue, street = self.canvas.calculate_location(event.x, event.y)
        action = self.action_var.get()
        if action == "move_karel":
            if avenue != self.karel.avenue or street != self.karel.street:
                self.karel.avenue = avenue
                self.karel.street = street
                self.canvas.redraw_karel()
        elif action == "add_beeper":
            apply_function(self.world.add_beeper)
        elif action == "remove_beeper":
            apply_function(self.world.remove_beeper)
        elif action == "reset_corner":
            apply_function(self.world.reset_corner)
        elif action == "paint_corner":
            apply_function(self.world.paint_corner, self.color_var.get())
        elif action == "add_wall":
            wall = self.canvas.find_nearest_wall(event.x, event.y, avenue, street)
            if wall:
                self.world.add_wall(wall)
                self.canvas.redraw_walls()
        elif action == "remove_wall":
            wall = self.canvas.find_nearest_wall(event.x, event.y, avenue, street)
            if wall:
                self.world.remove_wall(wall)
                self.canvas.redraw_walls()

    def save_world(self):
        default_worlds_path = os.path.join(os.path.dirname(__file__), "worlds")
        filename = asksaveasfilename(
            initialdir=default_worlds_path,
            title="Save Karel World",
            filetypes=[("Karel Worlds", "*.w")],
            parent=self.master,
        )
        if filename == "":
            return
        if not filename.endswith(".w"):
            filename += ".w"
        self.world.save_to_file(filename, self.karel)


if __name__ == "__main__":
    run_world_editor()



"""
This file defines the class definition of a Karel world.

The sub header comment defines important notes about the Karel
world file format.

General Notes About World Construction
- Streets run EAST-WEST (rows)
- Avenues run NORTH-SOUTH (columns)

World File Constraints:
- World file should specify one component per line in the format
  KEYWORD: PARAMETERS
- Any lines with no colon delimiter will be ignored
- The accepted KEYWORD, PARAMETER combinations are as follows:
    - Dimension: (num_avenues, num_streets)
    - Wall: (avenue, street); direction
    - Beeper: (avenue, street) count
    - Karel: (avenue, street); direction
    - Color: (avenue, street); color
    - Speed: delay
    - BeeperBag: num_beepers
- Multiple parameter values for the same keyword should be separated by a semicolon
- All numerical values (except delay) must be expressed as ints. The exception
  to this is that the number of beepers can also be INFINITY
- Any specified color values must be valid TKinter color strings, and are limited
  to the set of colors
- Direction is case-insensitive and can be one of the following values:
    - East
    - West
    - North
    - South

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""
import collections
import copy
import os
import re
import sys

from .karel_definitions import (
    COLOR_MAP,
    DIRECTIONS_MAP,
    DIRECTIONS_MAP_INVERSE,
    INFINITY,
    Direction,
    Wall,
)

INIT_SPEED = 50
VALID_WORLD_KEYWORDS = ["dimension", "wall", "beeper", "karel", "speed", "beeperbag", "color"]
KEYWORD_DELIM = ":"
PARAM_DELIM = ";"
DEFAULT_WORLD_FILE = "default_world.w"


class KarelWorld:
    def __init__(self, world_file):
        """
        Karel World constructor
        Parameters:
            world_file: filename of world file containing the initial state of Karel's world
        """
        self.world_name = world_file
        self._world_file = self.process_world(world_file)

        # Map of beeper locations to the count of beepers at that location
        self._beepers = collections.defaultdict(int)

        # Map of corner colors, defaults to None
        self._corner_colors = collections.defaultdict(lambda: "")

        # Set of Wall objects placed in the world
        self._walls = set()

        # Dimensions of the world
        self._num_streets = 1
        self._num_avenues = 1

        # Initial Karel state saved to enable world reset
        self._karel_starting_location = (1, 1)
        self._karel_starting_direction = Direction.EAST
        self._karel_starting_beeper_count = 0

        # Initial speed slider setting
        self._init_speed = INIT_SPEED

        # If a world file has been specified, load world details from the file
        if self._world_file:
            self.load_from_file()

        # Save initial beeper state to enable world reset
        self._init_beepers = copy.deepcopy(self._beepers)

    def __eq__(self, other):
        return (
            self.beepers == other.beepers
            and self.walls == other.walls
            and self.num_streets == other.num_streets
            and self.num_avenues == other.num_avenues
            and self.corner_colors == other.corner_colors
        )

    @staticmethod
    def process_world(world_file):
        """
        If no world_file is provided, use default world.
        Find world file that matches program name in the current worlds/ directory.
        If not found, search the provided default worlds directory.
        """
        default_worlds_path = os.path.join(os.path.dirname(__file__), "worlds")
        if not world_file:
            default_world = os.path.join(default_worlds_path, DEFAULT_WORLD_FILE)
            if os.path.isfile(default_world):
                print("Using default world...")
                return open(default_world)
            raise FileNotFoundError(
                "Default world cannot be found in: {}\n"
                "Please raise an issue on the stanfordkarel GitHub.".format(default_worlds_path)
            )

        if os.path.isfile(world_file):
            return open(world_file)

        for worlds_path in ("worlds", default_worlds_path):
            if os.path.isdir(worlds_path):
                full_world_path = os.path.join(worlds_path, world_file + ".w")
                if os.path.isfile(full_world_path):
                    return open(full_world_path)

        if not os.path.isdir("worlds"):
            print("Could not find worlds/ folder in current directory.\n")

        sys.tracebacklimit = 0
        raise FileNotFoundError(
            "The specified file was not one of the provided worlds.\n"
            "Please store custom worlds in a folder named worlds/, or use a world listed below:\n{}"
            "\nPass the default world as a parameter in run_karel_program().\n"
            "    e.g. run_karel_program('checkerboard_karel')".format(
                "\n".join(
                    [
                        (" " * 4) + os.path.splitext(world)[0]
                        for world in sorted(os.listdir(default_worlds_path))
                    ]
                )
            )
        )

    @property
    def karel_starting_location(self):
        return self._karel_starting_location

    @property
    def karel_starting_direction(self):
        return self._karel_starting_direction

    @property
    def karel_starting_beeper_count(self):
        return self._karel_starting_beeper_count

    @property
    def init_speed(self):
        return self._init_speed

    @property
    def num_streets(self):
        return self._num_streets

    @num_streets.setter
    def num_streets(self, val):
        self._num_streets = val

    @property
    def num_avenues(self):
        return self._num_avenues

    @num_avenues.setter
    def num_avenues(self, val):
        self._num_avenues = val

    @property
    def beepers(self):
        return self._beepers

    @property
    def corner_colors(self):
        return self._corner_colors

    @property
    def walls(self):
        return self._walls

    def load_from_file(self):
        def parse_parameters(keyword, param_str):
            params = {}
            for param in param_str.split(PARAM_DELIM):
                param = param.strip()

                # check to see if parameter encodes a location
                coordinate = re.match(r"\((\d+),\s*(\d+)\)", param)
                if coordinate:
                    # avenue, street
                    params["location"] = (int(coordinate.group(1)), int(coordinate.group(2)))
                    continue

                # check to see if the parameter is a direction value
                if param in DIRECTIONS_MAP:
                    params["direction"] = DIRECTIONS_MAP[param]

                # check to see if parameter encodes a numerical value or color string
                elif keyword == "color":
                    if param.title() not in COLOR_MAP:
                        raise ValueError(
                            "Error: {} is invalid parameter for {}.".format(param, keyword)
                        )
                    params["color"] = param.title()

                # handle the edge case where Karel has infinite beepers
                elif param in ("infinity", "infinite") and keyword == "beeperbag":
                    params["val"] = INFINITY

                # float values are only valid for the speed parameter.
                elif keyword == "speed":
                    try:
                        params["val"] = int(100 * float(param))
                    except ValueError:
                        raise ValueError(
                            "Error: {} is invalid parameter for {}.".format(param, keyword)
                        )

                # must be a digit then
                elif param.isdigit():
                    params["val"] = int(param)

                else:
                    raise ValueError(
                        "Error: {} is invalid parameter for {}.".format(param, keyword)
                    )

            return params

        for i, line in enumerate(self._world_file):
            # Ignore blank lines and lines with no comma delineator
            line = line.strip()
            if not line:
                continue

            if ":" not in line:
                print(
                    "Incorrectly formatted line - ignoring line {} of world file: {}".format(
                        i, line
                    )
                )
                continue

            keyword, param_str = line.lower().split(KEYWORD_DELIM)

            # only accept valid keywords as defined in world file spec
            # TODO: add error detection for keywords with insufficient parameters
            params = parse_parameters(keyword, param_str)

            # handle all different possible keyword cases
            if keyword == "dimension":
                # set world dimensions based on location values
                self._num_avenues, self._num_streets = params["location"]

            elif keyword == "wall":
                # build a wall at the specified location
                (avenue, street), direction = params["location"], params["direction"]
                self._walls.add(Wall(avenue, street, direction))

            elif keyword == "beeper":
                # add the specified number of beepers to the world
                self._beepers[params["location"]] += params["val"]

            elif keyword == "karel":
                # Give Karel initial state values
                self._karel_starting_location = params["location"]
                self._karel_starting_direction = params["direction"]

            elif keyword == "beeperbag":
                # Set Karel's initial beeper bag count
                self._karel_starting_beeper_count = params["val"]

            elif keyword == "speed":
                # Set delay speed of program execution
                self._init_speed = params["val"]

            elif keyword == "color":
                # Set corner color to be specified color
                self._corner_colors[params["location"]] = params["color"]

            else:
                print("Invalid keyword - ignoring line {} of world file: {}".format(i, line))

    def add_beeper(self, avenue, street):
        self._beepers[(avenue, street)] += 1

    def remove_beeper(self, avenue, street):
        if self._beepers[(avenue, street)] == 0:
            return
        self._beepers[(avenue, street)] -= 1

    def add_wall(self, wall):
        alt_wall = self.get_alt_wall(wall)
        if wall not in self._walls and alt_wall not in self._walls:
            self._walls.add(wall)

    def remove_wall(self, wall):
        alt_wall = self.get_alt_wall(wall)
        if wall in self._walls:
            self._walls.remove(wall)
        if alt_wall in self._walls:
            self._walls.remove(alt_wall)

    @staticmethod
    def get_alt_wall(wall):
        if wall.direction == Direction.NORTH:
            return Wall(wall.avenue, wall.street + 1, Direction.SOUTH)
        if wall.direction == Direction.SOUTH:
            return Wall(wall.avenue, wall.street - 1, Direction.NORTH)
        if wall.direction == Direction.EAST:
            return Wall(wall.avenue + 1, wall.street, Direction.WEST)
        if wall.direction == Direction.WEST:
            return Wall(wall.avenue - 1, wall.street, Direction.EAST)
        raise ValueError

    def paint_corner(self, avenue, street, color):
        self._corner_colors[(avenue, street)] = color

    def corner_color(self, avenue, street):
        return self._corner_colors[(avenue, street)]

    def reset_corner(self, avenue, street):
        self._beepers[(avenue, street)] = 0
        self._corner_colors[(avenue, street)] = ""

    def wall_exists(self, avenue, street, direction):
        wall = Wall(avenue, street, direction)
        return wall in self._walls

    def in_bounds(self, avenue, street):
        return 0 < avenue <= self._num_avenues and 0 < street <= self._num_streets

    def reset_world(self):
        """
        Reset initial state of beepers in the world
        """
        self._beepers = copy.deepcopy(self._init_beepers)
        self._corner_colors = collections.defaultdict(lambda: "")

    def reload_world(self, filename=None):
        """
        Reloads world using constructor.
        """
        self.__init__(filename)

    def save_to_file(self, filename, karel):
        with open(filename, "w") as f:
            # First, output dimensions of world
            f.write("Dimension: ({}, {})\n".format(self.num_avenues, self.num_streets))

            # Next, output all walls
            for wall in self._walls:
                f.write(
                    "Wall: ({}, {}); "
                    "{}\n".format(wall.avenue, wall.street, DIRECTIONS_MAP_INVERSE[wall.direction])
                )

            # Next, output all beepers
            for loc, count in self._beepers.items():
                f.write("Beeper: ({}, {}); {}\n".format(loc[0], loc[1], count))

            # Next, output all color information
            for loc, color in self._corner_colors.items():
                if color:
                    f.write("Color: ({}, {}); {}\n".format(loc[0], loc[1], color))

            # Next, output Karel information
            f.write(
                "Karel: ({}, {}); "
                "{}\n".format(karel.avenue, karel.street, DIRECTIONS_MAP_INVERSE[karel.direction])
            )

            # Finally, output beeperbag info
            beeper_output = karel.num_beepers if karel.num_beepers >= 0 else "INFINITY"
            f.write("BeeperBag: {}\n".format(beeper_output))



"""
This file defines the necessary functions and definitions that students must
import in order to be able to write a new Karel program. Any new Karel file
must include the following line:

	from stanfordkarel import *

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""

import sys
import tkinter as tk

from .karel import Karel
from .karel_application import KarelApplication


"""
The following function definitions are defined as stubs so that IDEs can recognize
the function definitions in student code. These names are re-bound upon program
execution to asscoiate their behavior to the one particular Karel object located
in a given world.
"""


def move():
    pass


def turn_left():
    pass


def put_beeper():
    pass


def pick_beeper():
    pass


def front_is_clear():
    pass


def front_is_blocked():
    pass


def left_is_clear():
    pass


def left_is_blocked():
    pass


def right_is_clear():
    pass


def right_is_blocked():
    pass


def beepers_present():
    pass


def no_beepers_present():
    pass


def beepers_in_bag():
    pass


def no_beepers_in_bag():
    pass


def facing_north():
    pass


def not_facing_north():
    pass


def facing_east():
    pass


def not_facing_east():
    pass


def facing_west():
    pass


def not_facing_west():
    pass


def facing_south():
    pass


def not_facing_south():
    pass


def paint_corner():
    pass


def corner_color_is():
    pass


# Defined constants for ease of use by students when coloring corners
RED = "red"
BLACK = "black"
CYAN = "cyan"
DARK_GRAY = "gray30"
GRAY = "gray55"
GREEN = "green"
LIGHT_GRAY = "gray80"
MAGENTA = "magenta3"
ORANGE = "orange"
PINK = "pink"
WHITE = "snow"
BLUE = "blue"
YELLOW = "yellow"
BLANK = ""


def run_karel_program(world_file=""):
    # Extract the name of the file the student is executing
    student_code_file = sys.argv[0]

    # Can uncomment this code to allow the file name to match the intended file
    # if world_file is None:
    #     base_filename = os.path.basename(student_code_file)
    #     world_file = os.path.splitext(base_filename)[0]

    # Create Karel and assign it to live in the newly created world
    karel = Karel(world_file)

    # Initialize root Tk Window and spawn Karel application
    root = tk.Tk()
    app = KarelApplication(karel, student_code_file, master=root)
    app.mainloop()



from stanfordkarel.stanfordkarel import (
    BLACK,
    BLANK,
    BLUE,
    CYAN,
    DARK_GRAY,
    GRAY,
    GREEN,
    LIGHT_GRAY,
    MAGENTA,
    ORANGE,
    PINK,
    RED,
    WHITE,
    YELLOW,
    beepers_in_bag,
    beepers_present,
    corner_color_is,
    facing_east,
    facing_north,
    facing_south,
    facing_west,
    front_is_blocked,
    front_is_clear,
    left_is_blocked,
    left_is_clear,
    move,
    no_beepers_in_bag,
    no_beepers_present,
    not_facing_east,
    not_facing_north,
    not_facing_south,
    not_facing_west,
    paint_corner,
    pick_beeper,
    put_beeper,
    right_is_blocked,
    right_is_clear,
    run_karel_program,
    turn_left,
)



"""
This file defines the GUI for running Karel programs.

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
Last Modified: 3/31/2020
"""

import importlib.util
import inspect
import os
import sys
import tkinter as tk
import traceback as tb
from time import sleep
from tkinter.filedialog import askopenfilename
from tkinter.messagebox import showwarning

from .karel_canvas import KarelCanvas
from .karel_definitions import DEFAULT_ICON, LIGHT_GREY, PAD_X, PAD_Y, KarelException


class StudentCode:
    """
    This process extracts a module from an arbitary file that contains student code.
    https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    """

    def __init__(self, code_file, karel=None):
        assert os.path.isfile(code_file), "{} could not be found.".format(code_file)
        self.module_name = os.path.basename(code_file)
        if self.module_name.endswith(".py"):
            self.module_name = os.path.splitext(self.module_name)[0]
        spec = importlib.util.spec_from_file_location(self.module_name, os.path.abspath(code_file))
        try:
            self.mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.mod)
        except Exception as e:
            # Handle syntax errors and only print location of error
            print("Syntax Error: {}".format(e))
            print("\n".join(tb.format_exc(limit=0).split("\n")[1:]))
            sys.exit()

        # Do not proceed if the student has not defined a main function.
        if not hasattr(self.mod, "main"):
            print("Couldn't find the main() function. Are you sure you have one?")
            sys.exit()

        if karel is not None:
            self.inject_namespace(karel)

    def __repr__(self):
        return inspect.getsource(self.mod)

    def inject_namespace(self, karel):
        """
		This function is responsible for doing some Python hackery
		that associates the generic commands the student wrote in their
		file with specific commands relating to the Karel object that exists
		in the world.
		"""
        functions_to_override = [
            "move",
            "turn_left",
            "pick_beeper",
            "put_beeper",
            "facing_north",
            "facing_south",
            "facing_east",
            "facing_west",
            "not_facing_north",
            "not_facing_south",
            "not_facing_east",
            "not_facing_west",
            "front_is_clear",
            "beepers_present",
            "no_beepers_present",
            "beepers_in_bag",
            "no_beepers_in_bag",
            "front_is_blocked",
            "left_is_blocked",
            "left_is_clear",
            "right_is_blocked",
            "right_is_clear",
            "paint_corner",
            "corner_color_is",
        ]
        for func in functions_to_override:
            setattr(self.mod, func, getattr(karel, func))


class KarelApplication(tk.Frame):
    def __init__(
        self,
        karel,
        code_file,
        master=None,
        window_width=800,
        window_height=600,
        canvas_width=600,
        canvas_height=400,
    ):
        # set window background to contrast white Karel canvas
        master.configure(background=LIGHT_GREY)

        # configure location of canvas to expand to fit window resizing
        master.rowconfigure(0, weight=1)
        master.columnconfigure(1, weight=1)

        # set master geometry
        master.geometry(str(window_width) + "x" + str(window_height))

        super().__init__(master, background=LIGHT_GREY)

        self.karel = karel
        self.world = karel.world
        self.student_code = StudentCode(code_file, self.karel)
        if not self.student_code.mod:
            master.destroy()
            return
        self.icon = DEFAULT_ICON
        self.window_width = window_width
        self.window_height = window_height
        self.canvas_width = canvas_width
        self.canvas_height = canvas_height
        self.master = master
        self.master.title(self.student_code.module_name)
        self.set_dock_icon()
        self.inject_decorator_namespace()
        self.grid(row=0, column=0)
        self.create_menubar()
        self.create_canvas()
        self.create_buttons()
        self.create_slider()
        self.create_status_label()

    def set_dock_icon(self):
        # make Karel dock icon image
        path = os.path.join(os.path.dirname(__file__), "icon.png")
        img = tk.Image("photo", file=path)
        self.master.tk.call("wm", "iconphoto", self.master._w, img)

    def create_menubar(self):
        menubar = tk.Menu(self.master)
        file_menu = tk.Menu(menubar, tearoff=False)
        menubar.add_cascade(label="File", menu=file_menu)
        file_menu.add_command(
            label="Exit", underline=1, command=self.master.quit, accelerator="Cmd+W"
        )
        iconmenu = tk.Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Select Icon", menu=iconmenu)
        iconmenu.add_command(label="Karel", command=lambda: self.set_icon("karel"))
        iconmenu.add_command(label="Simple", command=lambda: self.set_icon("simple"))

        self.bind_all("<Command-w>", self.quit)
        self.master.config(menu=menubar)

    def quit(self, event):
        del event
        sys.exit(0)

    def set_icon(self, icon):
        self.canvas.set_icon(icon)
        self.canvas.redraw_karel()

    def create_slider(self):
        """
		This method creates a frame containing three widgets:
		two labels on either side of a scale slider to control
		Karel execution speed.
		"""
        self.slider_frame = tk.Frame(self, bg=LIGHT_GREY)
        self.slider_frame.grid(row=3, column=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.fast_label = tk.Label(self.slider_frame, text="Fast", bg=LIGHT_GREY)
        self.fast_label.pack(side="right")

        self.slow_label = tk.Label(self.slider_frame, text="Slow", bg=LIGHT_GREY)
        self.slow_label.pack(side="left")

        self.speed = tk.IntVar()

        self.scale = tk.Scale(
            self.slider_frame, orient=tk.HORIZONTAL, variable=self.speed, showvalue=0
        )
        self.scale.set(self.world.init_speed)
        self.scale.pack()

    def create_canvas(self):
        """
		This method creates the canvas on which Karel and Karel's
		world are drawn.
		"""
        self.canvas = KarelCanvas(
            self.canvas_width, self.canvas_height, self.master, world=self.world, karel=self.karel
        )
        self.canvas.grid(column=1, row=0, sticky="NESW")
        self.canvas.bind("<Configure>", lambda t: self.canvas.redraw_all())

    def create_buttons(self):
        """
		This method creates the three buttons that appear on the left
		side of the screen. These buttons control the start of Karel
		execution, resetting Karel's state, and loading new worlds.
		"""
        self.program_control_button = tk.Button(
            self, highlightthickness=0, highlightbackground="white"
        )
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program
        self.program_control_button.grid(column=0, row=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.load_world_button = tk.Button(
            self, highlightthickness=0, text="Load World", command=self.load_world
        )
        self.load_world_button.grid(column=0, row=2, padx=PAD_X, pady=PAD_Y, sticky="ew")

    def create_status_label(self):
        """
		This function creates the status label at the bottom of the window.
		"""
        self.status_label = tk.Label(self.master, text="Welcome to Karel!", bg=LIGHT_GREY)
        self.status_label.grid(row=1, column=0, columnspan=2)

    def karel_action_decorator(self, karel_fn):
        def wrapper():
            # execute Karel function
            karel_fn()
            # redraw canvas with updated state of the world
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def beeper_action_decorator(self, karel_fn):
        def wrapper():
            # execute Karel function
            karel_fn()
            # redraw canvas with updated state of the world
            self.canvas.redraw_beepers()
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def corner_action_decorator(self, karel_fn):
        def wrapper(color):
            # execute Karel function
            karel_fn(color)
            # redraw canvas with updated state of the world
            self.canvas.redraw_corners()
            self.canvas.redraw_beepers()
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def inject_decorator_namespace(self):
        """
		This function is responsible for doing some Python hackery
		that associates the generic commands the student wrote in their
		file with specific commands relating to the Karel object that exists
		in the world.
		"""
        self.student_code.mod.turn_left = self.karel_action_decorator(self.karel.turn_left)
        self.student_code.mod.move = self.karel_action_decorator(self.karel.move)
        self.student_code.mod.pick_beeper = self.beeper_action_decorator(self.karel.pick_beeper)
        self.student_code.mod.put_beeper = self.beeper_action_decorator(self.karel.put_beeper)
        self.student_code.mod.paint_corner = self.corner_action_decorator(self.karel.paint_corner)

    def disable_buttons(self):
        self.program_control_button.configure(state="disabled")
        self.load_world_button.configure(state="disabled")

    def enable_buttons(self):
        self.program_control_button.configure(state="normal")
        self.load_world_button.configure(state="normal")

    def display_error_traceback(self, e):
        print("Traceback (most recent call last):")
        display_frames = []
        # walk through all the frames in stack trace at time of failure
        for frame, lineno in tb.walk_tb(e.__traceback__):
            frame_info = inspect.getframeinfo(frame)
            # get the name of the file corresponding to the current frame
            filename = frame_info.filename
            # Only display frames generated within the student's code
            if self.student_code.module_name + ".py" in filename:
                display_frames.append((frame, lineno))

        print(("".join(tb.format_list(tb.StackSummary.extract(display_frames)))).strip())
        print("{}: {}".format(type(e).__name__, str(e)))

    def run_program(self):
        # Error checking for existence of main function completed in prior file
        try:
            self.status_label.configure(text="Running...", fg="brown")
            self.disable_buttons()
            self.student_code.mod.main()
            self.status_label.configure(text="Finished running.", fg="green")

        except (KarelException, NameError) as e:
            # Generate popup window to let the user know their program crashed
            self.status_label.configure(
                text="Program crashed, check console for details.", fg="red"
            )
            self.display_error_traceback(e)
            self.update()
            showwarning("Karel Error", "Karel Crashed!\nCheck the terminal for more details.")

        finally:
            # Update program control button to force user
            # to reset world before running program again
            self.program_control_button["text"] = "Reset World"
            self.program_control_button["command"] = self.reset_world
            self.enable_buttons()

    def reset_world(self):
        self.karel.reset_state()
        self.world.reset_world()
        self.canvas.redraw_all()
        self.status_label.configure(text="Reset to initial state.", fg="black")
        # Once world has been reset, program control button resets to "run" mode
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program
        self.update()

    def load_world(self):
        default_worlds_path = os.path.join(os.path.dirname(__file__), "worlds")
        filename = askopenfilename(
            initialdir=default_worlds_path,
            title="Select Karel World",
            filetypes=[("Karel Worlds", "*.w")],
            parent=self.master,
        )
        # User hit cancel and did not select file, so leave world as-is
        if filename == "":
            return
        self.world.reload_world(filename=filename)
        self.karel.reset_state()
        self.canvas.redraw_all()
        # Reset speed slider
        self.scale.set(self.world.init_speed)
        self.status_label.configure(
            text="Loaded world from {}.".format(os.path.basename(filename)), fg="black"
        )

        # Make sure program control button is set to 'run' mode
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program



import setuptools

with open("README.md") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="stanfordkarel",
    version="0.2.2",
    author="Tyler Yep",
    author_email="tyep10@gmail.com",
    description="Official Stanford Karel library used in CS 106A",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/tyleryep/stanford-karel",
    packages=["stanfordkarel"],
    include_package_data=True,
    python_requires=">=3.5",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
This file defines the Karel ASCII program.

Original Author: Tyler Yep
License: MIT
Version: 1.0.0
Email: tyep@cs.stanford.edu
"""
from enum import Enum, unique

from .karel_definitions import Direction

CHAR_WIDTH = 6
HORIZONTAL, VERTICAL = "─", "│"
SPACING = 10


class Tile:
    def __init__(self, value="·"):
        self.value = value
        self.walls = []
        self.beepers = 0
        self.color = ""

    def __repr__(self):
        result = ""
        if self.value == "K" and self.beepers > 0:
            result += " <K> "
        elif self.beepers > 0:
            result += " <{}> ".format(self.beepers)
        elif self.color:
            result += " {} ".format(self.color[:3])
        else:
            result += "  {}  ".format(self.value)
        return result


@unique
class Color(Enum):
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def compare_output(first, second):
    """ Compares Karel Output and gets the results. """

    def create_two_column_string(col1, col2):
        """ col1 and col2 are Lists. """
        return map(lambda x: "{}{}{}".format(x[0], " " * SPACING, x[1]), zip(col1, col2))

    def symmetric_difference(a, b):
        extra_a, extra_b = {}, {}
        for k in a:
            if k not in b:
                extra_a[k] = a[k]
            elif a[k] - b[k] > 0:
                extra_a[k] = a[k] - b[k]
        for k in b:
            if k not in a:
                extra_b[k] = b[k]
            elif b[k] - a[k] > 0:
                extra_b[k] = b[k] - a[k]
        return extra_a, extra_b

    # if not two_column_output:
    # print(f"\n\nStudent output:\n{self}")
    # print(f"\nExpected output:\n{other}")

    this, that = str(first).split("\n"), str(second).split("\n")
    world_width = len(this[0])

    header1, header2 = "Student Output:", "Expected Output:"
    text_spacing = " " * (world_width - len(header1) + SPACING + 1)
    two_columns = create_two_column_string(this, that)
    output = "\n".join(two_columns)
    fancy_arrows = "{}❯{}❯{}❯ ".format(Color.RED.value, Color.YELLOW.value, Color.GREEN.value)

    result = "\n\n{} {}{}{}" "\n{}{}{}\n{}\n".format(
        fancy_arrows,
        Color.YELLOW.value,
        first.world.world_name,
        Color.END.value,
        header1,
        text_spacing,
        header2,
        output,
    )

    if first.avenue != second.avenue or first.street != second.street:
        result += (
            "Karel did not end up in the same location in both worlds:\n"
            "Student: {}\n"
            "Expected: {}\n\n".format((first.avenue, first.street), (second.avenue, second.street))
        )
    if first.world.beepers != second.world.beepers:
        extra_a, extra_b = symmetric_difference(first.world.beepers, second.world.beepers)
        result += (
            "Beepers do not match: "
            "(Only beepers that appear in one world but not the other are listed)\n"
            "Student: {}\n"
            "Expected: {}\n\n".format(extra_a, extra_b)
        )
    return result


def karel_ascii(world, karel_street, karel_avenue):
    """ Creates a Karel World in ASCII Art! """

    def tile_pair_has_wall(r, c, direction):
        """ Checks if the tile at r, c should have a wall by checking itself and its neighbors. """

        def tile_has_wall(r, c, direction):
            if 0 <= r < world.num_streets and 0 <= c < world.num_avenues:
                tile = world_arr[r][c]
                return direction in tile.walls
            return False

        if direction == Direction.SOUTH:
            return tile_has_wall(r, c, Direction.SOUTH) or tile_has_wall(r + 1, c, Direction.NORTH)
        if direction == Direction.NORTH:
            return tile_has_wall(r, c, Direction.NORTH) or tile_has_wall(r - 1, c, Direction.SOUTH)
        if direction == Direction.WEST:
            return tile_has_wall(r, c, Direction.WEST) or tile_has_wall(r, c - 1, Direction.EAST)
        if direction == Direction.EAST:
            return tile_has_wall(r, c, Direction.EAST) or tile_has_wall(r, c + 1, Direction.WEST)

    def get_next_line(r, c, next_block_start):
        """ Given a tile, figures out what the lower line of the ASCII art should be. """

        if tile_pair_has_wall(r, c, Direction.SOUTH):
            if (
                next_block_start == HORIZONTAL
                and tile_pair_has_wall(r, c, Direction.WEST)
                and tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "┼"
            elif (
                next_block_start == " "
                and tile_pair_has_wall(r, c, Direction.WEST)
                and tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "├"
            elif tile_pair_has_wall(r + 1, c, Direction.WEST):
                next_block_start = "┌"
            elif tile_pair_has_wall(r, c, Direction.WEST):
                next_block_start = "└"
            next_line = next_block_start + HORIZONTAL * (CHAR_WIDTH - 1)
            next_block_start = HORIZONTAL
        else:
            if (
                next_block_start == HORIZONTAL
                and tile_pair_has_wall(r, c, Direction.WEST)
                and tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "┤"
            elif next_block_start == HORIZONTAL and tile_pair_has_wall(r + 1, c, Direction.WEST):
                next_block_start = "┐"
            elif next_block_start == HORIZONTAL and tile_pair_has_wall(r, c, Direction.WEST):
                next_block_start = "┘"
            elif tile_pair_has_wall(r, c, Direction.WEST) and tile_pair_has_wall(
                r + 1, c, Direction.WEST
            ):
                next_block_start = VERTICAL

            next_line = next_block_start + " " * (CHAR_WIDTH - 1)
            next_block_start = " "
        return next_line, next_block_start

    world_arr = [[Tile() for _ in range(world.num_avenues)] for _ in range(world.num_streets)]

    world_arr[world.num_streets - karel_street][karel_avenue - 1].value = "K"
    for (avenue, street), count in world.beepers.items():
        world_arr[world.num_streets - street][avenue - 1].beepers = count

    for wall in world.walls:
        avenue, street, direction = wall.avenue, wall.street, wall.direction
        world_arr[world.num_streets - street][avenue - 1].walls.append(direction)

    for r in range(1, world.num_streets + 1):
        for c in range(1, world.num_avenues + 1):
            if world.corner_color(c, r):
                world_arr[world.num_streets - r][c - 1].color = world.corner_color(c, r)

    result = "┌{}┐\n".format(HORIZONTAL * (CHAR_WIDTH * world.num_avenues + 1))
    for r in range(world.num_streets):
        next_line = VERTICAL
        result += VERTICAL
        next_block_start = " "
        for c in range(world.num_avenues):
            tile = world_arr[r][c]
            line, next_block_start = get_next_line(r, c, next_block_start)
            next_line += line
            result += VERTICAL if tile_pair_has_wall(r, c, Direction.WEST) else " "
            result += str(tile)

        result += " {}\n".format(VERTICAL)
        if r == world.num_streets - 1:
            result += "└{}┘\n".format(HORIZONTAL * (CHAR_WIDTH * world.num_avenues + 1))
        else:
            result += "{} {}\n".format(next_line, VERTICAL)
    return result



"""
This file defines the Style Checker for Karel programs.

Original Author: Tyler Yep
Credits: Brahm Capoor
License: MIT
Version: 1.0.0
Email: tyep@cs.stanford.edu
"""
import ast
import inspect

import stanfordkarel

from .karel_application import StudentCode


def style_test(func):
    """ Decorator for printing out the results of a style test. """

    def success_func(*args, **kwargs):
        result = func(*args, **kwargs)
        if result:
            print("All good!")
        else:
            print("Error!")
        return result

    return success_func


class StyleChecker:
    """
    #TODO check variable name lengths
    #TODO some measure of decomposition
    #TODO check recursion
    """

    def __init__(self, code_file):
        self.student_code = StudentCode(code_file)
        self.module = self.student_code.mod
        self.module_lines = str(self.student_code).split("\n")
        self.function_list = [
            o[0] for o in inspect.getmembers(self.module) if inspect.isfunction(o[1])
        ]

    @staticmethod
    def print_status_message(message: str) -> None:
        print(message.ljust(64), end="")

    def check_style(self):
        print("\n\nStyle Tests for {}:".format(self.student_code.module_name))
        assert self.check_line_lengths()
        assert self.check_function_defs()
        assert self.assert_num_functions()
        assert self.check_recursion()
        assert self.check_naming()

    @style_test
    def check_recursion(self) -> bool:
        self.print_status_message("Checking for recursion...")
        return True

    @style_test
    def check_line_lengths(self, max_line_length=88) -> bool:
        self.print_status_message("Checking line lengths...")
        long_idxs = [i for i, line in enumerate(self.module_lines) if len(line) > max_line_length]
        if long_idxs:
            print("Lines {} are longer than {} characters.\n".format(long_idxs, max_line_length))
        return len(long_idxs) == 0

    @style_test
    def check_function_defs(self, min_name_length=5) -> bool:
        self.print_status_message("Checking function definitions...")
        ok = True
        seen_already = set()
        for f in self.function_list:
            if f in seen_already:
                print("There are multiple functions defined that are called {}".format(f))
                ok = False
            seen_already.add(f)
            if len(f) < min_name_length and f not in ("main", "move"):
                print("function {} has a pretty short name.".format(f))
                ok = False
        return ok

    @style_test
    def check_naming(self, min_name_length=5) -> bool:
        """. """
        self.print_status_message("Checking function and variable names...")
        stanfordkarel_names = dir(stanfordkarel)
        root = ast.parse(str(self.student_code))
        names = sorted({node.id for node in ast.walk(root) if isinstance(node, ast.Name)})
        filtered_names = [name for name in names if name not in stanfordkarel_names]

        short_names = [
            f for f in filtered_names if len(f) < min_name_length and f not in ("main", "_")
        ]
        if short_names:
            print("Functions {} have pretty short names.".format(short_names))
        return len(short_names) == 0

    @style_test
    def assert_num_functions(self, min_required=10) -> bool:
        """
        Check that *at least* `num` functions
        are present in the module.
        """
        self.print_status_message("Checking number of functions...")
        num_functions = len(self.function_list)
        if num_functions < min_required:
            print(
                "Expected at least {} functions, only found {}.".format(min_required, num_functions)
            )
        return num_functions >= min_required
